"""
    Selector related helper functions
"""
from operator import is_
from typing import Callable, TypeVar

import rx.operators as op
from rx import Observable, pipe
from rx.subject import ReplaySubject

T1 = TypeVar('T1')
T2 = TypeVar('T2')
Mapper = Callable[[T1], T2]


def select(selector: Mapper[T1, T2]
           ) -> Callable[[Observable], Observable]:
    """ Reactive operator that applies a selector
        and shares the result across multiple subscribers

        Args:
            selector: the selector function

        Returns:
            The reactive operator
    """
    return pipe(
        op.map(selector),
        op.distinct_until_changed(comparer=is_),
        op.multicast(subject=ReplaySubject(1)),
        op.ref_count(),
    )



from typing import Sequence, cast

from redux import Action, Reducer, handle_actions

from .action import ACTION_ADD_TODO, ACTION_TOGGLE_TODO, TodoPayload
from .state import TodoItem


def handle_add_todo(state: Sequence[TodoItem],
                    action: Action) -> Sequence[TodoItem]:
    payload = cast(TodoPayload, action.payload)
    return (*state, TodoItem(payload.id, payload.text, False))


def handle_toggle_todo(
        state: Sequence[TodoItem], action: Action) -> Sequence[TodoItem]:

    key = cast(int, action.payload)

    return [TodoItem(item.id, item.text, not item.completed)
            if item.id == key else item for item in state]


todos: Reducer = handle_actions(
    {ACTION_ADD_TODO: handle_add_todo, ACTION_TOGGLE_TODO: handle_toggle_todo}, ())



"""
    Implements epic specific functions
"""

from functools import partial
from inspect import getfullargspec
from logging import getLogger
from typing import Iterable, cast

from rx import Observable, merge
from rx.core.typing import Mapper

from .types import Epic

logger = getLogger(__name__)


def _wrapped_epic(epic: Mapper[Observable, Observable],
                  action_: Observable, _: Observable) -> Observable:
    return epic(action_)


def normalize_epic(epic: Epic) -> Epic:
    """Creates a callback for an epic that expects two arguments

    Args:
        epic: the epic

    Returns:
        the normalized epic
    """
    count = len(getfullargspec(epic)[0])
    assert count in (1, 2)
    return epic if count == 2 else cast(Epic, partial(_wrapped_epic, epic))


def _run_epic(action_: Observable, state_: Observable, epic: Epic) -> Observable:
    return epic(action_, state_)


def run_epic(action_: Observable, state_: Observable) -> Mapper[Epic, Observable]:
    """ Runs a single epic agains the given action and state observables

        Args:
            action_: the action observable
            state_: the state observable

        Returns:
            A callback function that will run the given epic on the observables
    """

    assert isinstance(action_, Observable)
    assert isinstance(state_, Observable)

    return partial(_run_epic, action_, state_)


def _combine_epics(
    norm_epics: Iterable[Epic],
    action_: Observable, state_: Observable
) -> Observable:
    """ Merges the epics into one

        Args:
            action_: the action observable
            state_: the state observable

        Returns:
            the merged epic
    """
    return merge(*map(run_epic(action_, state_), norm_epics))


def combine_epics(*epics: Iterable[Epic]) -> Epic:
    """ Combines a sequence of epics into one single epic by merging them

        Args:
            epics: the epics to merge

        Returns:
            The merged epic
    """
    return partial(_combine_epics, tuple(map(normalize_epic, epics)))



from typing import Iterable, MutableMapping, MutableSequence, Optional, Sequence
from unittest.case import TestCase

from rx import Observable, operators
from rx.subject import BehaviorSubject
from rx.testing import ReactiveTest, TestScheduler

from redux import ReduxRootStore, create_store, select

from .action import DECREMENT_ACTION, INCREMENT_ACTION
from .feature import create_counter_feature, select_counter_feature


class TestCounter(TestCase):

    def test_type(self):

        def reduce_to_list(dst: Iterable[int], src: int) -> Iterable:
            return (*dst, src)

        store = create_store()
        store.add_feature_module(create_counter_feature())

        result = BehaviorSubject(None)

        store_: Observable[ReduxRootStore] = store.as_observable()
        store_.pipe(
            operators.map(select_counter_feature),
            operators.reduce(reduce_to_list, ()),
            operators.first()
        ).subscribe(result)

        store.dispatch(INCREMENT_ACTION)
        store.dispatch(DECREMENT_ACTION)

        store.on_completed()

        assert result.value == (0, 1, 0)



from typing import Any
from rx import Observable, pipe
from rx.operators import do_action, filter, map, ignore_elements

from redux import (
    Epic,
    Reducer,
    ReduxFeatureModule,
    combine_epics,
    create_action,
    create_feature_module,
    handle_actions,
    of_init_feature,
    of_type,
    select_action_payload,
    select_feature,
    StateType,
    Action,
)

SAMPLE_FEATURE = "SAMPLE_FEATURE"

ADD_SAMPLE_ACTION = "ADD_SAMPLE_ACTION"

add_sample_action = create_action(ADD_SAMPLE_ACTION)
select_sample_feature_module = select_feature(SAMPLE_FEATURE)


def create_sample_feature() -> ReduxFeatureModule:
    """
        Constructs a new sample feature
    """

    def handle_sample_action(state: Any, action: Action) -> Any:
        return select_action_payload(action)

    sample_reducer = handle_actions({ADD_SAMPLE_ACTION: handle_sample_action})

    add_epic = pipe(of_type(ADD_SAMPLE_ACTION), ignore_elements(),)

    sample_epic = combine_epics(add_epic)

    return create_feature_module(SAMPLE_FEATURE, sample_reducer, sample_epic)



from typing import cast

from redux import Action, Reducer, create_action, handle_actions

from .action import ACTION_SET_VISIBILITY_FILTER, VisibilityFilters


def handle_visibility_filter(state: VisibilityFilters,
                             action: Action) -> VisibilityFilters:
    payload = cast(VisibilityFilters, action.payload)
    print('payload', payload)
    return payload


visibility_filter: Reducer = handle_actions(
    {ACTION_SET_VISIBILITY_FILTER: handle_visibility_filter}, VisibilityFilters.SHOW_ALL)



""" Action objects and constants """
from redux import create_action
from .constants import FEATURE_NAME


ACTION_INCREMENT = '%s Increment' % FEATURE_NAME
ACTION_DECREMENT = '%s Decrement' % FEATURE_NAME

DECREMENT_ACTION = create_action(ACTION_DECREMENT)(1)
INCREMENT_ACTION = create_action(ACTION_INCREMENT)(1)



from unittest.case import TestCase

from redux import Action

from .action import add_todo, toggle_todo
from .todo_reducer import todos


class TestTodoReducer(TestCase):

    def test_initial_state(self):
        result = todos(None, Action('sample', None))
        assert result == ()

    def test_add_todo(self):

        r1 = todos((), add_todo('todo1'))
        assert len(r1) == 1

        r2 = todos(r1, add_todo('todo2'))
        assert len(r1) == 1
        assert len(r2) == 2

    def test_toggle_todo(self):
        r1 = todos((), add_todo('todo1'))
        r2 = todos(r1, add_todo('todo2'))

        idx = len(r2) - 1
        id2 = r2[idx].id

        r3 = todos(r2, toggle_todo(id2))

        assert r3[idx].id == id2
        assert r3[idx].completed == True

        assert r2[idx].id == id2
        assert r2[idx].completed == False



"""
    Implementation of a Redux store with support for adding feature modules, dynamically.
    The store exposes a reactive API based on `RxPY <https://pypi.org/project/Rx/>`_.
"""

# Version of ReduxPy package
__version__ = "0.1.11"

from typing import Tuple

from ._internal.action import create_action, of_type, select_action_payload
from ._internal.epic import combine_epics
from ._internal.feature import (create_feature_module, of_init_feature,
                                select_feature)
from ._internal.reducer import combine_reducers, handle_actions
from ._internal.selectors import select
from ._internal.store import create_store
from ._internal.types import (Action, Epic, Reducer, ReduxFeatureModule,
                              ReduxRootStore, StateType)

__all__: Tuple[str, ...] = (
    'Action',
    'combine_epics',
    'combine_reducers',
    'create_action',
    'create_feature_module',
    'create_store',
    'Epic',
    'handle_actions',
    'of_init_feature',
    'of_type',
    'Reducer',
    'ReduxFeatureModule',
    'ReduxRootStore',
    'select_action_payload',
    'select_feature',
    'select',
    'StateType',
)



""" Exposes the feature creator """
from redux import ReduxFeatureModule, create_feature_module, select, select_feature

from .constants import FEATURE_NAME
from .reducer import todo_reducer

select_todos_feature = select_feature(FEATURE_NAME)


def create_todos_feature() -> ReduxFeatureModule:
    """Creates a new feature module

    Returns:
        ReduxFeatureModule -- the feature module
    """
    return create_feature_module(FEATURE_NAME, todo_reducer)



"""
    Implements feature specific functions
"""

from functools import partial
from logging import getLogger
from typing import Any, Callable, Iterable, Optional

import rx.operators as op
from rx import Observable, pipe
from rx.core.typing import Mapper, Predicate

from .action import is_by_selector, is_type, select_action_payload
from .constants import INIT_ACTION
from .types import (Action, Epic, Reducer, ReduxFeatureModule, ReduxRootState,
                    StateType)

logger = getLogger(__name__)


def has_payload(payload: Any) -> Predicate[Action]:
    """ Returns a function that checks if the action has a particular payload

        Args:
            payload: payload to test against

        Returns:
            Function to execute the check against an action
    """

    return is_by_selector(payload, select_action_payload)


def of_init_feature(
        identifier: str) -> Mapper[Observable, Observable]:
    """ Operator to test for the initialization action of a feature

        Args:
            identifier: the identifier of the feature

        Returns:
            Operator function that accepts init actions for the feature, once

    """
    return pipe(
        op.filter(is_type(INIT_ACTION)),
        op.filter(has_payload(identifier)),
        op.take(1),
        op.map(lambda x: identifier),
        op.do_action(logger.debug)
    )


def create_feature_module(
    identifier: str,
    reducer: Optional[Reducer] = None,
    epic: Optional[Epic] = None,
    dependencies: Iterable[ReduxFeatureModule] = (),
) -> ReduxFeatureModule:
    """ Constructs a new feature module descriptor

        Args:
            identifier: the identifier of the feature
            reducer: optional reducer
            epic: optional epic
            dependencies: optional dependencies on other features

        Returns:
            The feature module descriptor

    """
    return ReduxFeatureModule(identifier, reducer, epic, dependencies)


def _select_feature_by_id(
        identifier: str,
        initial_state: Optional[StateType],
        state: ReduxRootState) -> Optional[StateType]:
    """ Selector function that selects the feature state from the root state"""
    return state.get(identifier, initial_state)


def select_feature(
    identifier: str, initial_state: Optional[StateType] = None
) -> Callable[[ReduxRootState], Optional[StateType]]:
    """ Returns a function that returns the feature state from the root state

        Args:
            identifier: identifier of the feature
            initial_state: fallback state used if the feature state is not defined

        Returns:
            The selector function

    """
    return partial(_select_feature_by_id, identifier, initial_state)



""" Exposes the feature creator """
from redux import ReduxFeatureModule, create_feature_module, select_feature, select

from .constants import FEATURE_NAME
from .reducer import COUNTER_REDUCER

select_counter_feature = select_feature(FEATURE_NAME)


def create_counter_feature() -> ReduxFeatureModule:
    """Creates a new feature module

    Returns:
        ReduxFeatureModule -- the feature module
    """
    return create_feature_module(FEATURE_NAME, COUNTER_REDUCER)



from redux import Reducer, combine_reducers

from .todo_reducer import todos
from .visibility_reducer import visibility_filter

todo_reducer: Reducer = combine_reducers({
    'todos': todos,
    'visibility_filter': visibility_filter
})



"""
    Implements action specific functions
"""

from functools import partial
from typing import Any

import rx.operators as op
from rx import Observable
from rx.core.typing import Mapper, Predicate

from .types import Action, PayloadType


def create_action(type_name: str) -> Mapper[PayloadType, Action]:
    """ Creates a function that produces an action of the given type

        Args:
            type_name: type of the action

        Returns:
            A function that accepts the action payload and creates the action
    """
    return partial(Action, type_name)


def select_action_type(action: Action) -> str:
    """ Selects the type from the action

        Args:
            action: the action object

        Returns:
            the type of the action
    """
    return action.type


def select_action_payload(action: Action) -> PayloadType:
    """ Selects the payload from the action

        Args:
            action: the action object

        Returns:
            the payload of the action
    """
    return action.payload


def _is_by_selector(value: Any, selector: Mapper[Action, Any], action: Action) -> bool:
    """ Checks if selector on an action matches a value

        Args:
            action: the action object

        Returns:
            true if the selector result of the action matches, else false

    """

    return selector(action) is value


def is_by_selector(
    value: Any, selector: Mapper[Action, Any]
) -> Predicate[Action]:
    """ Returns a function that checks if the selector on an action equals a particular value

        Args:
            value: the value to compare against
            selector: the selector to use to extract the value from the action

        Returns:
            Function to execute the check against an action
    """
    return partial(_is_by_selector, value, selector)


def is_type(type_name) -> Predicate[Action]:
    """ Returns a function that checks if the action is of a particular type

        Args:
            type_name: type of the action to check for

        Returns:
            Function to execute the check against an action
    """

    return is_by_selector(type_name, select_action_type)


def of_type(
        type_name: str) -> Mapper[Observable, Observable]:
    """ Returns a reactive operator that filters for actions of the given type

        Args:
            type_name: type of the action to filter for

        Returns:
            The filter operator function
    """
    return op.filter(is_type(type_name))



from .feature import create_todos_feature, select_todos_feature
from .action import add_todo
from unittest.case import TestCase
from rx.operators import last
from rx.subject import BehaviorSubject

from redux import create_store


class TestTodoReducer(TestCase):

    def test_add_todo(self):

        store = create_store()

        store.add_feature_module(create_todos_feature())

        result = BehaviorSubject(None)

        store.as_observable().pipe(last()).subscribe(result)

        store.dispatch(add_todo('new todo'))
        store.dispatch(add_todo('another todo'))
        store.on_completed()

        feat = select_todos_feature(result.value)

        assert len(feat['todos']) == 2



"""
    Implements reducer related function
"""

from functools import partial
from typing import Iterable, Mapping, MutableMapping, Optional, Tuple

from .action import select_action_type
from .types import Action, Reducer, StateType


def _default_reducer(
        initial_state: Optional[StateType],
        state: StateType,
        _: Action) -> Optional[StateType]:
    #: default handling
    return state if state else initial_state


def default_reducer(initial_state: Optional[StateType]) -> Reducer:
    """ Creates a reducer that returns the original state or the default state.

        Args:
            inital_state: optional initial state used as a fallback

        Returns:
            A reducer that reduces the current state or the initial state as a fallback

    """
    return partial(_default_reducer, initial_state)


def _handle_actions_reducer(
        action_map: Mapping[str, Reducer],
        def_reducer: Reducer,
        state: StateType,
        action: Action) -> Optional[StateType]:
    #: dispatch
    return action_map.get(select_action_type(action),
                          def_reducer)(state, action)


def handle_actions(
    action_map: Mapping[str, Reducer], initial_state: Optional[StateType] = None,
) -> Reducer:
    """ Creates a new reducer from a mapping of action name to reducer for that action.

        Args:
            action_map: mapping from action name to reducer for that action
            initial_state: optional initial state used if no reducer matches

        Returns:
            A reducer function that handles the actions in the map

    """
    return partial(_handle_actions_reducer, {**action_map}, default_reducer(initial_state))


def _combine_reducers(items: Iterable[Tuple[str, Reducer]],
                      state: Mapping[str, StateType],
                      action: Action) -> Mapping[str, StateType]:
    """ Updates the state object from the reducer mappings. """
    result = state if state else {}
    mutable: Optional[MutableMapping[str, StateType]] = None
    for key, value in items:
        current = result.get(key)
        updated = value(current, action)
        if not current is updated:
            if not mutable:
                mutable = dict(result)
                result = mutable
            mutable[key] = updated
    return result


def combine_reducers(
        reducers: Mapping[str, Reducer]) -> Reducer[Mapping[str, StateType]]:
    """ Creates a new reducer from a mapping of reducers.

        Args:
            reducers: the mapping from state partition to reducer

        Returns:
            A reducer that dispatches actions against each of the mapped reducers

    """
    return partial(_combine_reducers, tuple(reducers.items()))



"""Setup script for redux-py"""

from os.path import abspath, dirname, join
from pathlib import Path
from typing import Sequence

from setuptools import find_packages, setup

# The directory containing this file
HERE = abspath(dirname(__file__))

# The text of the README file
with open(join(HERE, "README.md")) as fid:
    README = fid.read()


def read(fname: str) -> str:
    return open(Path(__file__).parent / fname).read()


def read_requirements(filename: str) -> Sequence[str]:
    return read(filename).splitlines()


# This call to setup() does all the work
setup(
    name="redux-py",
    version="0.1.11",
    description="Redux implementation for Python",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/Carsten-Leue/ReduxPY",
    author="Dr. Carsten Leue",
    author_email="carsten.leue@gmail.com",
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        "Programming Language :: Python :: 3",
        'Typing :: Typed'
    ],
    install_requires=read_requirements('requirements.txt'),
    tests_require=read_requirements('test-requirements.txt'),
    packages=find_packages(exclude=("tests",)),
    include_package_data=True,
    zip_safe=False
)



from typing import Any
from rx import Observable, pipe
from rx.operators import do_action, filter, map, ignore_elements

from redux import (
    Epic,
    Reducer,
    ReduxFeatureModule,
    combine_epics,
    create_action,
    create_feature_module,
    handle_actions,
    of_init_feature,
    of_type,
    select_action_payload,
    select_feature,
    StateType,
    Action,
)

INIT_FEATURE = "INIT_FEATURE"

ADD_INIT_ACTION = "ADD_INIT_ACTION"

add_init_action = create_action(ADD_INIT_ACTION)
select_init_feature_module = select_feature(INIT_FEATURE)


def create_init_feature() -> ReduxFeatureModule:
    """
        Constructs a new sample feature
    """

    def handle_init_action(state: Any, action: Action) -> Any:
        return select_action_payload(action)

    sample_reducer = handle_actions({ADD_INIT_ACTION: handle_init_action})

    add_epic = pipe(of_type(ADD_INIT_ACTION), ignore_elements(),)

    init_epic = pipe(
        of_init_feature(INIT_FEATURE), map(lambda x: add_init_action("init")),
    )

    sample_epic = combine_epics(add_epic, init_epic)

    return create_feature_module(INIT_FEATURE, sample_reducer, sample_epic)




"""
    Basic implementation of the store
"""

from logging import getLogger
from typing import Iterable, Mapping, MutableMapping, Optional, cast

import rx.operators as op
from rx import Observable, merge
from rx.subject import BehaviorSubject, Subject

from .action import create_action
from .constants import INIT_ACTION
from .epic import normalize_epic, run_epic
from .reducer import combine_reducers
from .types import (Action, Epic, Reducer, ReduxFeatureModule, ReduxRootState,
                    ReduxRootStore, StateType)

logger = getLogger(__name__)


def select_id(module: ReduxFeatureModule) -> str:
    """ Selects the ID from a module

        Args:
            module: the module

        Returns:
            The module identifier
    """
    return module.id


def select_dependencies(
        module: ReduxFeatureModule) -> Iterable[ReduxFeatureModule]:
    """ Selects the dependencies from a module

        Args:
            module: the module

        Returns:
            The module dependencies
    """
    return module.dependencies


def select_reducer(module: ReduxFeatureModule) -> Optional[Reducer]:
    """ Selects the reducer from a module

        Args:
            module: the module

        Returns:
            The module reducer
    """
    return module.reducer


def select_epic(module: ReduxFeatureModule) -> Optional[Epic]:
    """ Selects the epic from a module

        Args:
            module: the module

        Returns:
            The module epic
    """
    return module.epic


def has_reducer(module: ReduxFeatureModule) -> bool:
    """ Tests if a module defines a reducer

        Args:
            module: the module

        Returns:
            True if the module has a reducer, else False

    """
    return bool(select_reducer(module))


def identity_reducer(state: StateType, _: Action) -> StateType:
    """ Reducer function that returns the original state
    """
    return state


def reduce_reducers(
    dst: Mapping[str, Reducer], module: ReduxFeatureModule
) -> Mapping[str, Reducer]:
    """ Reduces the reducer on a module into a dictionary

        Args:
            dst: the target dictionary
            module: the module to reduce

        Returns:
            the reduced dictionary

    """
    reducer: Optional[Reducer] = select_reducer(module)
    return cast(Mapping[str, Reducer], {
                **dst, select_id(module): reducer}) if reducer else dst


def create_store(initial_state: Optional[ReduxRootState] = None) -> ReduxRootStore:  # pylint: disable=too-many-locals
    """ Constructs a new store that can handle feature modules.

        Args:
            initial_state: optional initial state of the store, will typically be the empty dict

        Returns:
            An implementation of the store
    """

    # current reducer
    reducer: Reducer = identity_reducer

    def replace_reducer(new_reducer: Reducer) -> None:
        """ Callback that replaces the current reducer

            Args:
                new_reducer: the new reducer

        """
        nonlocal reducer
        reducer = new_reducer

    # subject used to dispatch actions
    actions = Subject()

    # the shared action observable
    actions_ = actions.pipe(op.share())

    _dispatch = actions.on_next

    # our current state
    state = BehaviorSubject(initial_state if initial_state else {})

    # shutdown trigger
    done_ = Subject()

    # The set of known modules, to avoid cycles and duplicate registration
    modules: MutableMapping[str, ReduxFeatureModule] = {}

    # Sequence of added modules
    module_subject = Subject()

    # Subscribe to the resolved modules
    module_ = module_subject.pipe(op.distinct(select_id), op.share())

    # Build the reducers
    reducer_ = module_.pipe(
        op.filter(has_reducer),
        op.scan(reduce_reducers, {}),
        op.map(combine_reducers),
        op.map(replace_reducer),
    )

    # Build the epic
    epic_ = module_.pipe(
        op.map(select_epic),
        op.filter(bool),
        op.map(normalize_epic)
    )

    # Root epic that combines all of the incoming epics
    def root_epic(
        action_: Observable, state_: Observable
    ) -> Observable:
        """ Implementation of the root epic. If listens for new epics
            to come in and automatically subscribes.

            Args:
                action_: the action observable
                state_: the state observable

            Returns
                The observable of resulting actions
        """
        return epic_.pipe(
            op.flat_map(run_epic(action_, state_)),
            op.map(_dispatch)
        )

    # notifications about new feature states
    new_module_ = module_.pipe(
        op.map(select_id),
        op.map(create_action(INIT_ACTION)),
        op.map(_dispatch),
    )

    def _add_feature_module(module: ReduxFeatureModule):
        """ Registers a new feature module

            Args:
                module: the new feature module

        """
        module_id = select_id(module)
        if not module_id in modules:
            modules[module_id] = module
            for dep in select_dependencies(module):
                _add_feature_module(dep)
            module_subject.on_next(module)

    # all state
    internal_ = merge(root_epic(actions_, state), reducer_, new_module_).pipe(
        op.ignore_elements()
    )

    def _as_observable() -> Observable:
        """ Returns the state as an observable

            Returns:
                the observable
        """
        return state

    def _on_completed() -> None:
        """ Triggers the done event """
        done_.on_next(None)

    merge(actions_, internal_).pipe(
        op.map(lambda action: reducer(state.value, action)),
        op.take_until(done_),
    ).subscribe(state, logger.error)

    return ReduxRootStore(
        _as_observable, _dispatch, _add_feature_module, _dispatch, _on_completed
    )



""" Action objects and constants """
from enum import Enum
from itertools import count
from typing import Callable, Iterable, Iterator, NamedTuple

from redux import Action, create_action

from .constants import FEATURE_NAME

ACTION_ADD_TODO = '%s Add Todo' % FEATURE_NAME
ACTION_TOGGLE_TODO = '%s Toggle Todo' % FEATURE_NAME
ACTION_SET_VISIBILITY_FILTER = '%s Set Visibility Filter' % FEATURE_NAME


class VisibilityFilters(Enum):
    SHOW_ALL = 'SHOW_ALL'
    SHOW_COMPLETED = 'SHOW_COMPLETED'
    SHOW_ACTIVE = 'SHOW_ACTIVE'


class TodoPayload(NamedTuple):
    """ Todo payload """
    id: int
    text: str


_add_todo = create_action(ACTION_ADD_TODO)
_ids: Iterator[int] = iter(count())


# pylint: disable=unsubscriptable-object
def add_todo(text: str) -> Action:
    return _add_todo(TodoPayload(next(_ids), text))


# pylint: disable=unsubscriptable-object
set_visibility_filter: Callable[[VisibilityFilters], Action] = create_action(
    ACTION_SET_VISIBILITY_FILTER)

# pylint: disable=unsubscriptable-object
toggle_todo: Callable[[int], Action] = create_action(
    ACTION_TOGGLE_TODO)



from unittest.case import TestCase

from .action import (
    ACTION_ADD_TODO,
    ACTION_SET_VISIBILITY_FILTER,
    ACTION_TOGGLE_TODO,
    VisibilityFilters,
    add_todo,
    set_visibility_filter,
    toggle_todo,
)


class TestAction(TestCase):

    def test_add_todo(self):

        text = 'My Action'

        action = add_todo(text)
        assert action.type == ACTION_ADD_TODO
        assert action.payload.text == text

    def test_visibility_filter(self):

        fltr = VisibilityFilters.SHOW_ALL

        action = set_visibility_filter(fltr)
        assert action.type == ACTION_SET_VISIBILITY_FILTER
        assert action.payload == fltr

    def test_toggle_todo(self):

        item_id = 10

        action = toggle_todo(item_id)
        assert action.type == ACTION_TOGGLE_TODO
        assert action.payload == item_id



"""
    Module doc
"""
from typing import (Any, Callable, Iterable, Mapping, NamedTuple, Optional,
                    TypeVar)

from rx import Observable
from rx.core.typing import OnCompleted

StateType = TypeVar("StateType")

PayloadType = TypeVar("PayloadType")

ReduxRootState = Mapping[str, StateType]


class Action(NamedTuple):
    """ Action implementation that takes a payload """

    type: str
    """Identifier for the action, must be globally unique."""

    payload: Any
    """The action action payload"""


Epic = Callable[[Observable, Observable], Observable]

Reducer = Callable[[StateType, Action], StateType]


class ReduxFeatureModule(NamedTuple):
    """ Defines the feature module. The ID identifies the section in the state and
        is also used to globally discriminate features.

        After instantiating a feature store the store will fire an initialization action
        for that feature. Use :py:meth:`~redux.of_init_feature` to
        register for these initialization actions.
    """

    id: str
    """Identifier of the module, will also be used as a namespace into the state."""

    reducer: Optional[Reducer]
    """Reducer that handles module specific actions."""

    epic: Optional[Epic]
    """Epic that handles module specific asynchronous operations."""

    dependencies: Iterable['ReduxFeatureModule']
    """Dependencies on other feature modules"""


class ReduxRootStore(NamedTuple):
    """ Implementation of a store that manages sub-state as features. Features are added
        to the store automatically, when required by the select method.
    """

    as_observable: Callable[[], Observable]
    """Converts the store to an observable of state emissions"""

    dispatch: Callable[[Action], None]
    """Dispatches a single action to the store"""

    add_feature_module: Callable[[ReduxFeatureModule], None]
    """ Adds a new feature module """

    on_next: Callable[[Action], None]
    """ Alias for :py:meth:`~redux.ReduxRootStore.dispatch` """

    on_completed: OnCompleted
    """ Shuts down the store """



""" Implementation of the reducer """

from redux import Action, handle_actions

from .action import ACTION_DECREMENT, ACTION_INCREMENT


def handle_increment_action(state: int, action: Action) -> int:
    return state + 1


def handle_decrement_action(state: int, action: Action) -> int:
    return state - 1


COUNTER_REDUCER = handle_actions(
    {ACTION_INCREMENT: handle_increment_action, ACTION_DECREMENT: handle_decrement_action}, 0)
